/**
 * Copyright 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 */
package com.sunlabs.archsim.fleet.ships.fetchAndIssue;

import java.util.Collection;
import java.util.LinkedList;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.TriggerAction;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 * @author ib27688
 * A trigger-action that is at the core of 
 * the instruction issue/fetch unit component.
 */

class IssueComponent extends Component {
	
	protected LinkedList<Payload> instructions;
	
	IssueAction issueAction;
	Terminal out;
	Terminal ackOut;
	
	public IssueComponent(String name, Parameters params) {
		super(name, null);
		instructions= new LinkedList<Payload>();
		
		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				
				director.error("Component: " + name + ", id= " + getID() 
						+ ": Attribute " + pName + " with value " + sValue
						+ " is ignored");					
						
			} // for
		} // if params != null		
		
		buildIssueComponent();

	} // IssueComponent constructor
	
	public void buildIssueComponent() {
		
		// assemble the component, start with the action
		issueAction= new IssueAction(name+"IssueAction");
		addMember(issueAction);
		// now the terminals
		out= new OutputTerminal(FleetGlobal.OUT, this);
		ackOut= new InputTerminal(FleetGlobal.ACK_OUT, this);
		// attach the terminals
		issueAction.attach(out);
		issueAction.attach(ackOut);		
	}
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.OUT) || name.equals(out.getBaseAlias())) {
			return out;
		}
		else if (name.equals(FleetGlobal.ACK_OUT) || name.equals(ackOut.getBaseAlias())) {
			return ackOut;
		}
		else {
			return null;
		}
	} // getTerminal

	/**
	 * Add instructions to the pool of instructions to be issued
	 * Enable the issue command if necessary.
	 * @param param
	 */
	void addInstructions(Collection<Payload> newInstr) {
		if (newInstr != null && newInstr.size() > 0) {
			// add new instructions
			instructions.addAll(newInstr);
			// does the issue action need to be enabled?
			// enable if not
			if (!issueAction.isEnabled()) {
				issueAction.enableTriggerCommand.execute(instructions);
			}
		}
	} // add instructions

	/**
	 * Get the next instruction to be issued.
	 * This method can be revised to randomize the order of instruction issue;
	 * @return the next instruction to be issued.
	 */
	protected Payload nextInstruction() {
		Payload p;
		try {
			p= instructions.removeFirst();
		}
		catch (Exception e) {
			director.fatalError("IssueShip " 
					+ IssueComponent.this.getName() 
					+ ", " + IssueComponent.this.getID()
					+", attempting to issue an instruction "
					+" while having no instructions to issue");
			return null;
		} 
		return p;
	}
	
	protected class IssueAction extends TriggerAction {
		
		Function issueFn;
		
		public IssueAction(String name) {
			super(name);
			enabled= false;
			issueFn= new IssueFn(); 
			outputFunctions.put(FleetGlobal.OUT, issueFn);
		} // InstructionIssueAction constructor
		
		//issue the next available instruction
		// if this is the last instruction, disable the action
		protected class  IssueFn extends Function {
			
			public IssueFn(){
				super("IssueFn");
			} // IssueFn constructor
			
			public Object value(){
				Payload p= nextInstruction();
				// is this the last available instruction?
				// if so, disable the IssueAction, and, consequently,
				// the issue ship
				if (instructions.isEmpty()) {
					IssueAction.this.disableTriggerCommand.execute(null);
				}
				return p;
			} // value
		} // IssueFn
		
	} // class IssueAction
	
} // class IssueComponent

